//
//  Renderer.swift
//  Test2
//
//  Created by Caroline Begbie on 18/2/21.
//

import MetalKit

class Renderer: NSObject, MTKViewDelegate {
  
  static var device: MTLDevice!
  let commandQueue: MTLCommandQueue
  let pipeline: MTLComputePipelineState
  
  let setRange = 0..<1024
  var floatData = [Float]()
  var uintData = [CInt]()
  var floatDataBuffer: MTLBuffer!
  var uintDataBuffer: MTLBuffer!
  
  init?(metalView: MTKView) {
    guard let device = MTLCreateSystemDefaultDevice(),
          let commandQueue = device.makeCommandQueue() else { return nil }
    Renderer.device = device
    self.commandQueue = commandQueue
    
    let library = device.makeDefaultLibrary()!
    let sortFunction = library.makeFunction(name: "bitonicSort")!
    pipeline = try! device.makeComputePipelineState(function: sortFunction)
    
    super.init()
    metalView.delegate = self
    metalView.framebufferOnly = false
    metalView.clearColor = MTLClearColor(red: 0.2, green: 0.2,
                                         blue: 0.2, alpha: 1)
    
    for value in stride(from: Float(setRange.upperBound-1), to: Float(setRange.lowerBound-1), by: -1.0) {
      floatData.append(value)
    }
    for value in stride(from: setRange.upperBound-1, to: setRange.lowerBound-1, by: -1) {
      uintData.append(CInt(value))
    }
    floatDataBuffer = device.makeBuffer(bytes: &floatData,
                                        length: MemoryLayout<Float>.stride * floatData.count)!
    floatDataBuffer.label = "floatDataBuffer"
    
    uintDataBuffer = device.makeBuffer(bytes: &uintData,
                                       length: MemoryLayout<CInt>.stride * uintData.count)!
    uintDataBuffer.label = "uintDataBuffer"

    print("before")
    print(floatData)
    print(uintData)

    runSort()

    let dataPointer = floatDataBuffer.contents().assumingMemoryBound(to: Float.self)
    let dataBufferPointer = UnsafeMutableBufferPointer(start: dataPointer, count: floatData.count)
    floatData = Array.init(dataBufferPointer)

    let dataPointer2 = uintDataBuffer.contents().assumingMemoryBound(to: CInt.self)
    let dataBufferPointer2 = UnsafeMutableBufferPointer(start: dataPointer2, count: uintData.count)
    uintData = Array.init(dataBufferPointer2)

    print("After")
    print(floatData)
    print(uintData)


  }
  
  func runSort() {
    guard let logn = Int(exactly: log2(Double(floatData.count))) else {
      fatalError("data.count is not a power of 2")
    }
    for p in 0..<logn {
      for q in 0..<p+1 {
        
        var n1 = p
        var n2 = q
        guard let commandbuffer = commandQueue.makeCommandBuffer(),
              let encoder = commandbuffer.makeComputeCommandEncoder() else { return }
        
        encoder.setComputePipelineState(pipeline)
        encoder.setBuffer(floatDataBuffer, offset: 0, index: 0)
        encoder.setBuffer(uintDataBuffer, offset: 0, index: 1)
        encoder.setBytes(&n1, length: MemoryLayout<CInt>.stride, index: 2)
        encoder.setBytes(&n2, length: MemoryLayout<CInt>.stride, index: 3)
        let threadgroupsPerGrid = MTLSize(width: floatData.count, height: 1, depth: 1)
        let threadsPerThreadgroup = MTLSize(width: pipeline.threadExecutionWidth, height: 1, depth: 1)
        encoder.dispatchThreads(threadgroupsPerGrid, threadsPerThreadgroup: threadsPerThreadgroup)
        encoder.endEncoding()
        commandbuffer.commit()
        commandbuffer.waitUntilCompleted()
      }
    }
  }
  
  func mtkView(_ view: MTKView, drawableSizeWillChange size: CGSize) {
    
  }
  
  func draw(in view: MTKView) {
    print("Draw")
  }
}
